@@ -79,6 +79,11 @@
 #define XBUTTON2 2
 #endif
 
+#ifndef VK_XBUTTON1
+#define VK_XBUTTON1 5
+#define VK_XBUTTON2 6
+#endif
+
 #ifndef MK_XBUTTON1
 #define MK_XBUTTON1 32
 #define MK_XBUTTON2 64
@@ -107,6 +112,9 @@ static gboolean gdk_event_dispatch (GSource     *source,
 				    GSourceFunc  callback,
 				    gpointer     user_data);
 
+static void append_event (GdkDisplay *display,
+			  GdkEvent   *event);
+
 /* Private variable declarations
  */
 
@@ -154,7 +162,6 @@ static HKL latin_locale = NULL;
 #endif
 
 static gboolean in_ime_composition = FALSE;
-static gboolean resizing = FALSE;
 static UINT     resize_timer;
 
 static int debug_indent = 0;
@@ -221,11 +228,24 @@ _gdk_win32_get_next_tick (gulong suggested_tick)
   if (suggested_tick == 0)
     suggested_tick = GetTickCount ();
   if (suggested_tick <= cur_tick)
-    return ++cur_tick;
+    return cur_tick;
   else
     return cur_tick = suggested_tick;
 }
 
+static void
+generate_focus_event (GdkWindow *window,
+		      gboolean   in)
+{
+  GdkEvent *event;
+
+  event = gdk_event_new (GDK_FOCUS_CHANGE);
+  event->focus_change.window = window;
+  event->focus_change.in = in;
+
+  append_event (gdk_drawable_get_display (window), event);
+}
+
 static LRESULT 
 inner_window_procedure (HWND   hwnd,
 			UINT   message,
@@ -528,13 +548,13 @@ gdk_pointer_grab (GdkWindow    *window,
     hcursor = NULL;
   else if ((hcursor = CopyCursor (cursor_private->hcursor)) == NULL)
     WIN32_API_FAILED ("CopyCursor");
-#if 0
+
   return_val = _gdk_input_grab_pointer (window,
 					owner_events,
 					event_mask,
 					confine_to,
 					time);
-#endif
+
   if (return_val == GDK_GRAB_SUCCESS)
     {
       if (!GDK_WINDOW_DESTROYED (window))
@@ -610,9 +630,8 @@ gdk_display_pointer_ungrab (GdkDisplay *display,
   GDK_NOTE (EVENTS, g_print ("%sgdk_display_pointer_ungrab%s",
 			     (debug_indent > 0 ? "\n" : ""),
 			     (debug_indent == 0 ? "\n" : "")));
-#if 0
+
   _gdk_input_ungrab_pointer (time);
-#endif
 
   if (GetCapture () != NULL)
     ReleaseCapture ();
@@ -713,6 +732,8 @@ gdk_keyboard_grab (GdkWindow *window,
 		   gboolean   owner_events,
 		   guint32    time)
 {
+  GdkWindow *real_focus_window, *grab_focus_window;
+
   gint return_val;
   
   g_return_val_if_fail (window != NULL, 0);
@@ -730,7 +751,31 @@ gdk_keyboard_grab (GdkWindow *window,
     return_val = GDK_GRAB_ALREADY_GRABBED;
 
   if (return_val == GDK_GRAB_SUCCESS)
-    k_grab_window = window;
+    {
+      k_grab_window = window;
+
+      if (!k_grab_owner_events)
+	{
+	  real_focus_window = gdk_win32_handle_table_lookup ((GdkNativeWindow) GetFocus ());
+	  if (real_focus_window)
+	    real_focus_window = gdk_window_get_toplevel (real_focus_window);
+	  grab_focus_window = gdk_window_get_toplevel (k_grab_window);
+	  if (real_focus_window != grab_focus_window)
+	    {
+	      /* Generate events for focus change from the window that really
+	       * has focus to the grabber.
+	       */
+	      if (real_focus_window && !GDK_WINDOW_DESTROYED (real_focus_window)
+		  && (((GdkWindowObject *) real_focus_window)->event_mask
+		      & GDK_FOCUS_CHANGE_MASK))
+		generate_focus_event (real_focus_window, FALSE);
+
+	      if (((GdkWindowObject *) grab_focus_window)->event_mask 
+		  & GDK_FOCUS_CHANGE_MASK)
+		generate_focus_event (grab_focus_window, TRUE);
+	    }
+	}
+    }
   
   return return_val;
 }
@@ -739,10 +784,39 @@ void
 gdk_display_keyboard_ungrab (GdkDisplay *display,
                              guint32 time)
 {
+  GdkWindow *real_focus_window, *grab_focus_window;
+
   g_return_if_fail (display == gdk_display_get_default ());
 
   GDK_NOTE (EVENTS, g_print ("gdk_keyboard_ungrab\n"));
 
+  if (k_grab_window && !k_grab_owner_events)
+    {
+      real_focus_window = gdk_win32_handle_table_lookup ((GdkNativeWindow) GetFocus ());
+      if (real_focus_window)
+	real_focus_window = gdk_window_get_toplevel (real_focus_window);
+      if (!GDK_WINDOW_DESTROYED (k_grab_window))
+	grab_focus_window = gdk_window_get_toplevel (k_grab_window);
+      else
+	grab_focus_window = NULL;
+      if (real_focus_window != grab_focus_window)
+	{
+	  /* Generate events for focus change from grabber to the window that
+	   * really has focus. Important for example if a new window is created
+	   * while focus is grabbed.
+	   */
+	  if (grab_focus_window
+	      && (((GdkWindowObject *) grab_focus_window)->event_mask
+		  & GDK_FOCUS_CHANGE_MASK))
+	    generate_focus_event (grab_focus_window, FALSE);
+
+	  if (real_focus_window && !GDK_WINDOW_DESTROYED (real_focus_window)
+	      && (((GdkWindowObject *) real_focus_window)->event_mask
+		  & GDK_FOCUS_CHANGE_MASK))
+	    generate_focus_event (real_focus_window, TRUE);
+	}
+    }
+
   k_grab_window = NULL;
 }
 
@@ -827,6 +901,17 @@ build_key_event_state (GdkEvent *event,
   if (key_state[VK_CAPITAL] & 0x01)
     event->key.state |= GDK_LOCK_MASK;
 
+  if (key_state[VK_LBUTTON] & 0x80)
+    event->key.state |= GDK_BUTTON1_MASK;
+  if (key_state[VK_MBUTTON] & 0x80)
+    event->key.state |= GDK_BUTTON2_MASK;
+  if (key_state[VK_RBUTTON] & 0x80)
+    event->key.state |= GDK_BUTTON3_MASK;
+  if (key_state[VK_XBUTTON1] & 0x80)
+    event->key.state |= GDK_BUTTON4_MASK;
+  if (key_state[VK_XBUTTON2] & 0x80)
+    event->key.state |= GDK_BUTTON5_MASK;
+
   /* Win9x doesn't distinguish between left and right Control and Alt
    * in the keyboard state as returned by GetKeyboardState(), so we
    * have to punt, and accept either Control + either Alt to be AltGr.
@@ -1541,6 +1626,9 @@ translate_mouse_coords (GdkWindow *window1,
   msg->lParam = MAKELPARAM (pt.x, pt.y);
 }
 
+/* The check_extended flag controls whether to check if the windows want
+ * events from extended input devices and if the message should be skipped
+ * because an extended input device is active */
 static gboolean
 propagate (GdkWindow  **window,
 	   MSG         *msg,
@@ -1548,13 +1636,24 @@ propagate (GdkWindow  **window,
 	   gboolean     grab_owner_events,
 	   gint	        grab_mask,
 	   gboolean   (*doesnt_want_it) (gint mask,
-					 MSG *msg))
+					 MSG *msg),
+	   gboolean    	check_extended)
 {
   gboolean in_propagation = FALSE;
 
   if (grab_window != NULL && !grab_owner_events)
     {
       /* Event source is grabbed with owner_events FALSE */
+
+      /* See if the event should be ignored because an extended input device
+       * is used */
+      if (check_extended &&
+	  ((GdkWindowObject *) grab_window)->extension_events != 0 &&
+	  _gdk_input_ignore_core)
+	{
+	  GDK_NOTE (EVENTS, g_print (" (ignored for grabber)"));
+	  return FALSE;
+	}
       if ((*doesnt_want_it) (grab_mask, msg))
 	{
 	  GDK_NOTE (EVENTS, g_print (" (grabber doesn't want it)"));
@@ -1569,7 +1668,16 @@ propagate (GdkWindow  **window,
     }
   while (TRUE)
     {
-     if ((*doesnt_want_it) (((GdkWindowObject *) *window)->event_mask, msg))
+      /* See if the event should be ignored because an extended input device
+       * is used */
+      if (check_extended &&
+	  ((GdkWindowObject *) *window)->extension_events != 0 &&
+	  _gdk_input_ignore_core)
+	{
+	  GDK_NOTE (EVENTS, g_print (" (ignored)"));
+	  return FALSE;
+	}
+      if ((*doesnt_want_it) (((GdkWindowObject *) *window)->event_mask, msg))
 	{
 	  /* Owner doesn't want it, propagate to parent. */
 	  GdkWindow *parent = gdk_window_get_parent (*window);
@@ -1579,6 +1687,16 @@ propagate (GdkWindow  **window,
 	      if (grab_window != NULL)
 		{
 		  /* Event source is grabbed with owner_events TRUE */
+
+		  /* See if the event should be ignored because an extended
+		   * input device is used */
+		  if (check_extended &&
+		      ((GdkWindowObject *) grab_window)->extension_events != 0 &&
+		      _gdk_input_ignore_core)
+		    {
+		      GDK_NOTE (EVENTS, g_print (" (ignored for grabber)"));
+		      return FALSE;
+		    }
 		  if ((*doesnt_want_it) (grab_mask, msg))
 		    {
 		      /* Grabber doesn't want it either */
@@ -1669,23 +1787,29 @@ handle_configure_event (MSG       *msg,
 			GdkWindow *window)
 {
   RECT client_rect;
+  POINT point;
 
   GetClientRect (msg->hwnd, &client_rect);
-  
+  point.x = client_rect.left; /* always 0 */
+  point.y = client_rect.top;
+  /* top level windows need screen coords */
+  if (gdk_window_get_parent (window) == _gdk_parent_root)
+    {
+      ClientToScreen (msg->hwnd, &point);
+      point.x += _gdk_offset_x;
+      point.y += _gdk_offset_y;
+    }
+
   GDK_WINDOW_IMPL_WIN32 (((GdkWindowObject *) window)->impl)->width = client_rect.right - client_rect.left;
   GDK_WINDOW_IMPL_WIN32 (((GdkWindowObject *) window)->impl)->height = client_rect.bottom - client_rect.top;
   
-  ((GdkWindowObject *) window)->x = client_rect.left;
-  ((GdkWindowObject *) window)->y = client_rect.top;
+  ((GdkWindowObject *) window)->x = point.x;
+  ((GdkWindowObject *) window)->y = point.y;
   
   if (((GdkWindowObject *) window)->event_mask & GDK_STRUCTURE_MASK)
     {
-      POINT point;
       GdkEvent *event = gdk_event_new (GDK_CONFIGURE);
 
-      point.x = point.y = 0;
-      ClientToScreen (msg->hwnd, &point);
-      
       event->configure.window = window;
 
       event->configure.width = client_rect.right - client_rect.left;
@@ -1694,12 +1818,6 @@ handle_configure_event (MSG       *msg,
       event->configure.x = point.x;
       event->configure.y = point.y;
 
-      if (gdk_window_get_parent (window) == _gdk_parent_root)
-	{
-	  event->configure.x += _gdk_offset_x;
-	  event->configure.y += _gdk_offset_y;
-	}
-
       append_event (gdk_drawable_get_display (window), event);
     }
 }
@@ -2006,10 +2124,18 @@ resize_timer_proc (HWND     hwnd,
 		   UINT     id,
 		   DWORD    time)
 {
-  if (resizing)
+  if (_sizemove_in_progress)
     handle_stuff_while_moving_or_resizing ();
 }
 
+static void
+handle_display_change (void)
+{
+  _gdk_monitor_init ();
+  _gdk_root_window_size_init ();
+  g_signal_emit_by_name (_gdk_screen, "size_changed");
+}
+
 static gboolean
 gdk_event_translate (GdkDisplay *display,
 		     MSG        *msg,
@@ -2150,16 +2276,9 @@ gdk_event_translate (GdkDisplay *display,
 
       assign_object (&window, new_window);
 
-      if (((GdkWindowObject *) window)->extension_events != 0 &&
-	  _gdk_input_ignore_core)
-	{
-	  GDK_NOTE (EVENTS, g_print (" (ignored)"));
-	  goto done;
-	}
-
       if (!propagate (&window, msg,
 		      p_grab_window, p_grab_owner_events, p_grab_mask,
-		      doesnt_want_scroll))
+		      doesnt_want_scroll, TRUE))
 	goto done;
 
       if (GDK_WINDOW_DESTROYED (window))
@@ -2301,7 +2420,7 @@ gdk_event_translate (GdkDisplay *display,
 
       if (!propagate (&window, msg,
 		      k_grab_window, k_grab_owner_events, GDK_ALL_EVENTS_MASK,
-		      doesnt_want_key))
+		      doesnt_want_key, FALSE))
 	break;
 
       if (GDK_WINDOW_DESTROYED (window))
@@ -2383,7 +2502,7 @@ gdk_event_translate (GdkDisplay *display,
 
       if (!propagate (&window, msg,
 		      k_grab_window, k_grab_owner_events, GDK_ALL_EVENTS_MASK,
-		      doesnt_want_char))
+		      doesnt_want_char, FALSE))
 	break;
 
       if (GDK_WINDOW_DESTROYED (window))
@@ -2461,16 +2580,9 @@ gdk_event_translate (GdkDisplay *display,
 	    synthesize_crossing_events (window, GDK_CROSSING_NORMAL, msg);
 	}
 
-      if (((GdkWindowObject *) window)->extension_events != 0 &&
-	  _gdk_input_ignore_core)
-	{
-	  GDK_NOTE (EVENTS, g_print (" (ignored)"));
-	  break;
-	}
-
       if (!propagate (&window, msg,
 		      p_grab_window, p_grab_owner_events, p_grab_mask,
-		      doesnt_want_button_press))
+		      doesnt_want_button_press, TRUE))
 	break;
 
       if (GDK_WINDOW_DESTROYED (window))
@@ -2550,16 +2662,18 @@ gdk_event_translate (GdkDisplay *display,
 	    synthesize_crossing_events (window, GDK_CROSSING_NORMAL, msg);
 	}
 
+#if 0
       if (((GdkWindowObject *) window)->extension_events != 0 &&
 	  _gdk_input_ignore_core)
 	{
 	  GDK_NOTE (EVENTS, g_print (" (ignored)"));
 	  break;
 	}
+#endif
 
       if (!propagate (&window, msg,
 		      p_grab_window, p_grab_owner_events, p_grab_mask,
-		      doesnt_want_button_release))
+		      doesnt_want_button_release, TRUE))
 	{
 	}
       else if (!GDK_WINDOW_DESTROYED (window))
@@ -2622,16 +2736,9 @@ gdk_event_translate (GdkDisplay *display,
 	    synthesize_crossing_events (window, GDK_CROSSING_NORMAL, msg);
 	}
 
-      if (((GdkWindowObject *) window)->extension_events != 0 &&
-	  _gdk_input_ignore_core)
-	{
-	  GDK_NOTE (EVENTS, g_print (" (ignored)"));
-	  break;
-	}
-
       if (!propagate (&window, msg,
 		      p_grab_window, p_grab_owner_events, p_grab_mask,
-		      doesnt_want_button_motion))
+		      doesnt_want_button_motion, TRUE))
 	break;
 
       if (GDK_WINDOW_DESTROYED (window))
@@ -2729,16 +2836,9 @@ gdk_event_translate (GdkDisplay *display,
 	  assign_object (&window, new_window);
 	}
 
-      if (((GdkWindowObject *) window)->extension_events != 0 &&
-	  _gdk_input_ignore_core)
-	{
-	  GDK_NOTE (EVENTS, g_print (" (ignored)"));
-	  break;
-	}
-
       if (!propagate (&window, msg,
 		      p_grab_window, p_grab_owner_events, p_grab_mask,
-		      doesnt_want_scroll))
+		      doesnt_want_scroll, TRUE))
 	break;
 
       if (GDK_WINDOW_DESTROYED (window))
@@ -2804,7 +2904,7 @@ gdk_event_translate (GdkDisplay *display,
 
     case WM_SETFOCUS:
     case WM_KILLFOCUS:
-      if (p_grab_window != NULL && !p_grab_owner_events && !(p_grab_mask & GDK_FOCUS_CHANGE_MASK))
+      if (k_grab_window != NULL && !k_grab_owner_events)
 	break;
 
       if (!(((GdkWindowObject *) window)->event_mask & GDK_FOCUS_CHANGE_MASK))
@@ -2813,12 +2913,7 @@ gdk_event_translate (GdkDisplay *display,
       if (GDK_WINDOW_DESTROYED (window))
 	break;
 
-      event = gdk_event_new (GDK_FOCUS_CHANGE);
-      event->focus_change.window = window;
-      event->focus_change.in = (msg->message == WM_SETFOCUS);
-
-      append_event (display, event);
-
+      generate_focus_event (window, (msg->message == WM_SETFOCUS));
       return_val = TRUE;
       break;
 
@@ -2946,12 +3041,12 @@ gdk_event_translate (GdkDisplay *display,
       break;
 
     case WM_ENTERSIZEMOVE:
-      resizing = TRUE;
+      _sizemove_in_progress = TRUE;
       resize_timer = SetTimer (NULL, 0, 20, resize_timer_proc);
       break;
 
     case WM_EXITSIZEMOVE:
-      resizing = FALSE;
+      _sizemove_in_progress = FALSE;
       KillTimer (NULL, resize_timer);
       break;
 
@@ -2959,28 +3054,34 @@ gdk_event_translate (GdkDisplay *display,
       /* Once we've entered the moving or sizing modal loop, we won't
        * return to the main loop until we're done sizing or moving.
        */
-      if (resizing &&
+      if (_sizemove_in_progress &&
 	 GDK_WINDOW_TYPE (window) != GDK_WINDOW_CHILD &&
 	 !GDK_WINDOW_DESTROYED (window))
 	{
 	  RECT client_rect;
+	  POINT point;
 
 	  GetClientRect (msg->hwnd, &client_rect);
+	  point.x = client_rect.left; /* always 0 */
+	  point.y = client_rect.top;
+	  /* top level windows need screen coords */
+	  if (gdk_window_get_parent (window) == _gdk_parent_root)
+	    {
+	      ClientToScreen (msg->hwnd, &point);
+	      point.x += _gdk_offset_x;
+	      point.y += _gdk_offset_y;
+	    }
   
 	  GDK_WINDOW_IMPL_WIN32 (((GdkWindowObject *) window)->impl)->width = client_rect.right - client_rect.left;
 	  GDK_WINDOW_IMPL_WIN32 (((GdkWindowObject *) window)->impl)->height = client_rect.bottom - client_rect.top;
 	  
-	  ((GdkWindowObject *) window)->x = client_rect.left;
-	  ((GdkWindowObject *) window)->y = client_rect.top;
-	  
+	  ((GdkWindowObject *) window)->x = point.x;
+	  ((GdkWindowObject *) window)->y = point.y;
+
 	  if (((GdkWindowObject *) window)->event_mask & GDK_STRUCTURE_MASK)
 	    {
-	      POINT point;
 	      GdkEvent *event = gdk_event_new (GDK_CONFIGURE);
 	      
-	      point.x = point.y = 0;
-	      ClientToScreen (msg->hwnd, &point);
-	      
 	      event->configure.window = window;
 	      
 	      event->configure.width = client_rect.right - client_rect.left;
@@ -2989,12 +3090,6 @@ gdk_event_translate (GdkDisplay *display,
 	      event->configure.x = point.x;
 	      event->configure.y = point.y;
 	      
-	      if (gdk_window_get_parent (window) == _gdk_parent_root)
-		{
-		  event->configure.x += _gdk_offset_x;
-		  event->configure.y += _gdk_offset_y;
-		}
-	      
 	      if (((GdkWindowObject *) window)->resize_count > 1)
 		((GdkWindowObject *) window)->resize_count -= 1;
 	      
@@ -3256,6 +3351,11 @@ gdk_event_translate (GdkDisplay *display,
       return_val = TRUE;
       break;
 
+    case WM_DISPLAYCHANGE:
+      handle_display_change ();
+      break;
+      
+
 #ifdef HAVE_WINTAB
       /* Handle WINTAB events here, as we know that gdkinput.c will
        * use the fixed WT_DEFBASE as lcMsgBase, and we thus can use the
@@ -3281,6 +3381,7 @@ gdk_event_translate (GdkDisplay *display,
 
       event = gdk_event_new (GDK_NOTHING);
       event->any.window = window;
+      g_object_ref (window);
       if (_gdk_input_other_event (event, msg, window))
 	append_event (display, event);
       else
