@@ -294,7 +294,15 @@ gtk_cell_view_style_set (GtkWidget *widget,
 static void
 gtk_cell_view_finalize (GObject *object)
 {
-  gtk_cell_view_cell_layout_clear (GTK_CELL_LAYOUT (object));
+  GtkCellView *cellview = GTK_CELL_VIEW (object);
+
+  gtk_cell_view_cell_layout_clear (GTK_CELL_LAYOUT (cellview));
+
+  if (cellview->priv->model)
+     g_object_unref (cellview->priv->model);
+
+  if (cellview->priv->displayed_row)
+     gtk_tree_row_reference_free (cellview->priv->displayed_row);
 
   if (G_OBJECT_CLASS (parent_class)->finalize)
     (* G_OBJECT_CLASS (parent_class)->finalize) (object);
@@ -383,7 +391,7 @@ gtk_cell_view_size_allocate (GtkWidget     *widget,
       if (!info->cell->visible)
         continue;
 
-      info->real_width = info->requested_width + (info->expand?extra_space:0);
+      info->real_width = info->requested_width + (info->expand ? extra_space : 0);
     }
 
   /* iterate list for PACK_END cells */
@@ -397,7 +405,7 @@ gtk_cell_view_size_allocate (GtkWidget     *widget,
       if (!info->cell->visible)
         continue;
 
-      info->real_width = info->requested_width + (info->expand?extra_space:0);
+      info->real_width = info->requested_width + (info->expand ? extra_space : 0);
     }
 }
 
@@ -408,6 +416,7 @@ gtk_cell_view_expose (GtkWidget      *widget,
   GList *i;
   GtkCellView *cellview;
   GdkRectangle area;
+  GtkCellRendererState state;
   gboolean rtl = (gtk_widget_get_direction(widget) == GTK_TEXT_DIR_RTL);
 
   cellview = GTK_CELL_VIEW (widget);
@@ -450,6 +459,11 @@ gtk_cell_view_expose (GtkWidget      *widget,
   area.x = widget->allocation.x + (rtl ? widget->allocation.width : 0); 
   area.y = widget->allocation.y;
 
+  if (GTK_WIDGET_STATE (widget) == GTK_STATE_PRELIGHT)
+    state = GTK_CELL_RENDERER_PRELIT;
+  else
+    state = 0;
+      
   /* PACK_START */
   for (i = cellview->priv->cell_list; i; i = i->next)
     {
@@ -469,7 +483,7 @@ gtk_cell_view_expose (GtkWidget      *widget,
                                 event->window,
                                 widget,
                                 /* FIXME! */
-                                &area, &area, &event->area, 0);
+                                &area, &area, &event->area, state);
 
       if (!rtl)                                           
          area.x += info->real_width;
@@ -496,7 +510,7 @@ gtk_cell_view_expose (GtkWidget      *widget,
                                 widget->window,
                                 widget,
                                 /* FIXME ! */
-                                &area, &area, &event->area, 0);
+                                &area, &area, &event->area, state);
       if (rtl)
          area.x += info->real_width;
     }
@@ -529,7 +543,7 @@ gtk_cell_view_set_cell_data (GtkCellView *cellview)
   GtkTreePath *path;
 
   g_return_if_fail (cellview->priv->displayed_row != NULL);
-
+  
   path = gtk_tree_row_reference_get_path (cellview->priv->displayed_row);
   gtk_tree_model_get_iter (cellview->priv->model, &iter, path);
   gtk_tree_path_free (path);
@@ -634,7 +648,6 @@ gtk_cell_view_cell_layout_add_attribute (GtkCellLayout   *layout,
 static void
 gtk_cell_view_cell_layout_clear (GtkCellLayout *layout)
 {
-  GList *i;
   GtkCellView *cellview = GTK_CELL_VIEW (layout);
 
   g_return_if_fail (GTK_IS_CELL_VIEW (cellview));
@@ -691,17 +704,18 @@ gtk_cell_view_cell_layout_clear_attributes (GtkCellLayout   *layout,
   g_return_if_fail (GTK_IS_CELL_RENDERER (renderer));
 
   info = gtk_cell_view_get_cell_info (cellview, renderer);
-  g_return_if_fail (info != NULL);
-
-  list = info->attributes;
-  while (list && list->next)
+  if (info != NULL)
     {
-      g_free (list->data);
-      list = list->next->next;
+      list = info->attributes;
+      while (list && list->next)
+	{
+	  g_free (list->data);
+	  list = list->next->next;
+	}
+      
+      g_slist_free (info->attributes);
+      info->attributes = NULL;
     }
-
-  g_slist_free (info->attributes);
-  info->attributes = NULL;
 }
 
 static void
@@ -932,6 +946,7 @@ gtk_cell_view_get_size_of_row (GtkCellView    *cell_view,
                                GtkRequisition *requisition)
 {
   GtkTreeRowReference *tmp;
+  GtkRequisition req;
 
   g_return_val_if_fail (GTK_IS_CELL_VIEW (cell_view), FALSE);
   g_return_val_if_fail (path != NULL, FALSE);
@@ -946,6 +961,9 @@ gtk_cell_view_get_size_of_row (GtkCellView    *cell_view,
   gtk_tree_row_reference_free (cell_view->priv->displayed_row);
   cell_view->priv->displayed_row = tmp;
 
+  /* restore actual size info */
+  gtk_cell_view_size_request (GTK_WIDGET (cell_view), &req);
+
   return TRUE;
 }
 
