@@ -17,6 +17,7 @@
  * Boston, MA 02111-1307, USA.
  */
 
+#include <config.h>
 #include <string.h>
 #include <gobject/gvaluecollector.h>
 #include "gtktreemodel.h"
@@ -1866,12 +1867,12 @@ gtk_list_store_move (GtkListStore *store,
       if (a == iter->user_data)
         goto free_paths_and_out;
     }
-  else if (before)
+  else if (before && !position)
     {
       if (iter->user_data == store->tail)
         goto free_paths_and_out;
     }
-  else
+  else if (!position)
     {
       if (iter->user_data == store->root)
         goto free_paths_and_out;
@@ -1919,8 +1920,16 @@ gtk_list_store_move (GtkListStore *store,
     }
   else if (!a && before)
     {
-      G_SLIST (store->tail)->next = G_SLIST (iter->user_data);
-      G_SLIST (iter->user_data)->next = NULL;
+      if (position)
+	{
+	  G_SLIST (iter->user_data)->next = G_SLIST (store->root);
+	  store->root = G_SLIST (iter->user_data);
+	} 
+      else 
+	{
+	  G_SLIST (store->tail)->next = G_SLIST (iter->user_data);
+	  G_SLIST (iter->user_data)->next = NULL;
+	}
     }
 
   /* update tail if needed */
@@ -2094,7 +2103,8 @@ gtk_list_store_sort (GtkListStore *list_store)
   if (list_store->length <= 1)
     return;
 
-  g_assert (GTK_LIST_STORE_IS_SORTED (list_store));
+  if (!GTK_LIST_STORE_IS_SORTED (list_store))
+    return;
 
   list = G_SLIST (list_store->root);
 
@@ -2352,19 +2362,22 @@ gtk_list_store_set_sort_column_id (GtkTreeSortable  *sortable,
       (list_store->order == order))
     return;
 
-  if (sort_column_id != GTK_TREE_SORTABLE_DEFAULT_SORT_COLUMN_ID)
+  if (sort_column_id != -2)
     {
-      GtkTreeDataSortHeader *header = NULL;
-
-      header = _gtk_tree_data_list_get_header (list_store->sort_list, sort_column_id);
-
-      /* We want to make sure that we have a function */
-      g_return_if_fail (header != NULL);
-      g_return_if_fail (header->func != NULL);
-    }
-  else
-    {
-      g_return_if_fail (list_store->default_sort_func != NULL);
+      if (sort_column_id != GTK_TREE_SORTABLE_DEFAULT_SORT_COLUMN_ID)
+	{
+	  GtkTreeDataSortHeader *header = NULL;
+	  
+	  header = _gtk_tree_data_list_get_header (list_store->sort_list, sort_column_id);
+	  
+	  /* We want to make sure that we have a function */
+	  g_return_if_fail (header != NULL);
+	  g_return_if_fail (header->func != NULL);
+	}
+      else
+	{
+	  g_return_if_fail (list_store->default_sort_func != NULL);
+	}
     }
 
 
