@@ -24,6 +24,7 @@
  * GTK+ at ftp://ftp.gtk.org/pub/gtk/. 
  */
 
+#include <config.h>
 #include "gtktextutil.h"
 #include "gtkintl.h"
 #include "gtkmenuitem.h"
@@ -104,7 +105,7 @@ _gtk_text_util_append_special_char_menuitems (GtkMenuShell              *menushe
       
       menuitem = gtk_menu_item_new_with_mnemonic (_(bidi_menu_entries[i].label));
       g_object_set_data (G_OBJECT (menuitem), "gtk-unicode-menu-entry",
-                         &bidi_menu_entries[i]);
+                         (gpointer)&bidi_menu_entries[i]);
       
       g_signal_connect_data (menuitem, "activate",
                              G_CALLBACK (activate_cb),
